@@ -4,43 +4,45 @@ from git.config import GitConfigParser
 from io import BytesIO
 import weakref
 
-__all__ = ('sm_section', 'sm_name', 'mkhead', 'unbare_repo', 'find_first_remote_branch',
+
+# typing -----------------------------------------------------------------------
+
+from typing import Any, Sequence, TYPE_CHECKING, Union
+
+from git.types import PathLike
+
+if TYPE_CHECKING:
+    from .base import Submodule
+    from weakref import ReferenceType
+    from git.repo import Repo
+    from git.refs import Head
+    from git import Remote
+    from git.refs import RemoteReference
+
+
+__all__ = ('sm_section', 'sm_name', 'mkhead', 'find_first_remote_branch',
            'SubmoduleConfigParser')
 
 #{ Utilities
 
 
-def sm_section(name):
+def sm_section(name: str) -> str:
     """:return: section title used in .gitmodules configuration file"""
-    return 'submodule "%s"' % name
+    return f'submodule "{name}"'
 
 
-def sm_name(section):
+def sm_name(section: str) -> str:
     """:return: name of the submodule as parsed from the section name"""
     section = section.strip()
     return section[11:-1]
 
 
-def mkhead(repo, path):
+def mkhead(repo: 'Repo', path: PathLike) -> 'Head':
     """:return: New branch/head instance"""
     return git.Head(repo, git.Head.to_full_path(path))
 
 
-def unbare_repo(func):
-    """Methods with this decorator raise InvalidGitRepositoryError if they
-    encounter a bare repository"""
-
-    def wrapper(self, *args, **kwargs):
-        if self.repo.bare:
-            raise InvalidGitRepositoryError("Method '%s' cannot operate on bare repositories" % func.__name__)
-        # END bare method
-        return func(self, *args, **kwargs)
-    # END wrapper
-    wrapper.__name__ = func.__name__
-    return wrapper
-
-
-def find_first_remote_branch(remotes, branch_name):
+def find_first_remote_branch(remotes: Sequence['Remote'], branch_name: str) -> 'RemoteReference':
     """Find the remote branch matching the name of the given branch or raise InvalidGitRepositoryError"""
     for remote in remotes:
         try:
@@ -49,7 +51,7 @@ def find_first_remote_branch(remotes, branch_name):
             continue
         # END exception handling
     # END for remote
-    raise InvalidGitRepositoryError("Didn't find remote branch %r in any of the given remotes", branch_name)
+    raise InvalidGitRepositoryError("Didn't find remote branch '%r' in any of the given remotes" % branch_name)
 
 #} END utilities
 
@@ -67,19 +69,19 @@ class SubmoduleConfigParser(GitConfigParser):
     Please note that no mutating method will work in bare mode
     """
 
-    def __init__(self, *args, **kwargs):
-        self._smref = None
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        self._smref: Union['ReferenceType[Submodule]', None] = None
         self._index = None
         self._auto_write = True
         super(SubmoduleConfigParser, self).__init__(*args, **kwargs)
 
     #{ Interface
-    def set_submodule(self, submodule):
+    def set_submodule(self, submodule: 'Submodule') -> None:
         """Set this instance's submodule. It must be called before
         the first write operation begins"""
         self._smref = weakref.ref(submodule)
 
-    def flush_to_index(self):
+    def flush_to_index(self) -> None:
         """Flush changes in our configuration file to the index"""
         assert self._smref is not None
         # should always have a file here
@@ -98,8 +100,8 @@ class SubmoduleConfigParser(GitConfigParser):
     #} END interface
 
     #{ Overridden Methods
-    def write(self):
-        rval = super(SubmoduleConfigParser, self).write()
+    def write(self) -> None:
+        rval: None = super(SubmoduleConfigParser, self).write()
         self.flush_to_index()
         return rval
     # END overridden methods
