@@ -5,30 +5,51 @@ from git.exc import GitCommandError
 from .symbolic import SymbolicReference
 from .reference import Reference
 
+# typinng ---------------------------------------------------
+
+from typing import Any, Sequence, Union, TYPE_CHECKING
+
+from git.types import PathLike, Commit_ish
+
+if TYPE_CHECKING:
+    from git.repo import Repo
+    from git.objects import Commit
+    from git.refs import RemoteReference
+
+# -------------------------------------------------------------------
+
 __all__ = ["HEAD", "Head"]
 
 
+def strip_quotes(string):
+    if string.startswith('"') and string.endswith('"'):
+        return string[1:-1]
+    return string
+
+
 class HEAD(SymbolicReference):
 
     """Special case of a Symbolic Reference as it represents the repository's
     HEAD reference."""
     _HEAD_NAME = 'HEAD'
     _ORIG_HEAD_NAME = 'ORIG_HEAD'
-    __slots__ = tuple()
+    __slots__ = ()
 
-    def __init__(self, repo, path=_HEAD_NAME):
+    def __init__(self, repo: 'Repo', path: PathLike = _HEAD_NAME):
         if path != self._HEAD_NAME:
             raise ValueError("HEAD instance must point to %r, got %r" % (self._HEAD_NAME, path))
         super(HEAD, self).__init__(repo, path)
+        self.commit: 'Commit'
 
-    def orig_head(self):
+    def orig_head(self) -> SymbolicReference:
         """
         :return: SymbolicReference pointing at the ORIG_HEAD, which is maintained
             to contain the previous value of HEAD"""
         return SymbolicReference(self.repo, self._ORIG_HEAD_NAME)
 
-    def reset(self, commit='HEAD', index=True, working_tree=False,
-              paths=None, **kwargs):
+    def reset(self, commit: Union[Commit_ish, SymbolicReference, str] = 'HEAD',
+              index: bool = True, working_tree: bool = False,
+              paths: Union[PathLike, Sequence[PathLike], None] = None, **kwargs: Any) -> 'HEAD':
         """Reset our HEAD to the given commit optionally synchronizing
         the index and working tree. The reference we refer to will be set to
         commit as well.
@@ -54,8 +75,8 @@ class HEAD(SymbolicReference):
             Additional arguments passed to git-reset.
 
         :return: self"""
+        mode: Union[str, None]
         mode = "--soft"
-        add_arg = None
         if index:
             mode = "--mixed"
 
@@ -73,12 +94,8 @@ class HEAD(SymbolicReference):
 
         # END working tree handling
 
-        if paths:
-            add_arg = "--"
-        # END nicely separate paths from rest
-
         try:
-            self.repo.git.reset(mode, commit, add_arg, paths, **kwargs)
+            self.repo.git.reset(mode, commit, '--', paths, **kwargs)
         except GitCommandError as e:
             # git nowadays may use 1 as status to indicate there are still unstaged
             # modifications after the reset
@@ -112,7 +129,7 @@ class Head(Reference):
     k_config_remote_ref = "merge"           # branch to merge from remote
 
     @classmethod
-    def delete(cls, repo, *heads, **kwargs):
+    def delete(cls, repo: 'Repo', *heads: 'Head', **kwargs: Any):
         """Delete the given heads
 
         :param force:
@@ -125,7 +142,7 @@ class Head(Reference):
             flag = "-D"
         repo.git.branch(flag, *heads)
 
-    def set_tracking_branch(self, remote_reference):
+    def set_tracking_branch(self, remote_reference: 'RemoteReference') -> 'Head':
         """
         Configure this branch to track the given remote reference. This will alter
             this branch's configuration accordingly.
@@ -138,29 +155,26 @@ class Head(Reference):
             raise ValueError("Incorrect parameter type: %r" % remote_reference)
         # END handle type
 
-        writer = self.config_writer()
-        if remote_reference is None:
-            writer.remove_option(self.k_config_remote)
-            writer.remove_option(self.k_config_remote_ref)
-            if len(writer.options()) == 0:
-                writer.remove_section()
-            # END handle remove section
-        else:
-            writer.set_value(self.k_config_remote, remote_reference.remote_name)
-            writer.set_value(self.k_config_remote_ref, Head.to_full_path(remote_reference.remote_head))
-        # END handle ref value
-        writer.release()
+        with self.config_writer() as writer:
+            if remote_reference is None:
+                writer.remove_option(self.k_config_remote)
+                writer.remove_option(self.k_config_remote_ref)
+                if len(writer.options()) == 0:
+                    writer.remove_section()
+            else:
+                writer.set_value(self.k_config_remote, remote_reference.remote_name)
+                writer.set_value(self.k_config_remote_ref, Head.to_full_path(remote_reference.remote_head))
 
         return self
 
-    def tracking_branch(self):
+    def tracking_branch(self) -> Union['RemoteReference', None]:
         """
         :return: The remote_reference we are tracking, or None if we are
             not a tracking branch"""
         from .remote import RemoteReference
         reader = self.config_reader()
         if reader.has_option(self.k_config_remote) and reader.has_option(self.k_config_remote_ref):
-            ref = Head(self.repo, Head.to_full_path(reader.get_value(self.k_config_remote_ref)))
+            ref = Head(self.repo, Head.to_full_path(strip_quotes(reader.get_value(self.k_config_remote_ref))))
             remote_refpath = RemoteReference.to_full_path(join_path(reader.get_value(self.k_config_remote), ref.name))
             return RemoteReference(self.repo, remote_refpath)
         # END handle have tracking branch
@@ -168,7 +182,7 @@ class Head(Reference):
         # we are not a tracking branch
         return None
 
-    def rename(self, new_path, force=False):
+    def rename(self, new_path: PathLike, force: bool = False) -> 'Head':
         """Rename self to a new path
 
         :param new_path:
@@ -189,7 +203,7 @@ class Head(Reference):
         self.path = "%s/%s" % (self._common_path_default, new_path)
         return self
 
-    def checkout(self, force=False, **kwargs):
+    def checkout(self, force: bool = False, **kwargs: Any):
         """Checkout this head by setting the HEAD to this reference, by updating the index
         to reflect the tree we point to and by updating the working tree to reflect
         the latest index.
@@ -207,6 +221,8 @@ class Head(Reference):
         :return:
             The active branch after the checkout operation, usually self unless
             a new branch has been created.
+            If there is no active branch, as the HEAD is now detached, the HEAD
+            reference will be returned instead.
 
         :note:
             By default it is only allowed to checkout heads - everything else
@@ -217,11 +233,12 @@ class Head(Reference):
             kwargs.pop('f')
 
         self.repo.git.checkout(self, **kwargs)
+        if self.repo.head.is_detached:
+            return self.repo.head
         return self.repo.active_branch
 
-    #{ Configruation
-
-    def _config_parser(self, read_only):
+    #{ Configuration
+    def _config_parser(self, read_only: bool) -> SectionConstraint:
         if read_only:
             parser = self.repo.config_reader()
         else:
@@ -230,15 +247,15 @@ class Head(Reference):
 
         return SectionConstraint(parser, 'branch "%s"' % self.name)
 
-    def config_reader(self):
+    def config_reader(self) -> SectionConstraint:
         """
         :return: A configuration parser instance constrained to only read
             this instance's values"""
         return self._config_parser(read_only=True)
 
-    def config_writer(self):
+    def config_writer(self) -> SectionConstraint:
         """
-        :return: A configuration writer instance with read-and write acccess
+        :return: A configuration writer instance with read-and write access
             to options of this head"""
         return self._config_parser(read_only=False)
 
