@@ -13,7 +13,7 @@ from git.compat import (defenc, force_text)
 from git.exc import GitCommandError
 from git.util import (
     LazyMixin,
-    Iterable,
+    IterableObj,
     IterableList,
     RemoteProgress,
     CallableRemoteProgress,
@@ -36,20 +36,27 @@ from .refs import (
 
 # typing-------------------------------------------------------
 
-from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, TYPE_CHECKING, Union, cast, overload
+from typing import (Any, Callable, Dict, Iterator, List, NoReturn, Optional, Sequence,  # NOQA[TC002]
+                    TYPE_CHECKING, Type, Union, overload)
 
-from git.types import PathLike, Literal, TBD
+from git.types import PathLike, Literal, TBD, TypeGuard, Commit_ish     # NOQA[TC002]
 
 if TYPE_CHECKING:
     from git.repo.base import Repo
-    from git.objects.commit import Commit
-    from git.objects.blob import Blob
-    from git.objects.tree import Tree
-    from git.objects.tag import TagObject
+    from git.objects.submodule.base import UpdateProgress
+    # from git.objects.commit import Commit
+    # from git.objects import Blob, Tree, TagObject
+
+flagKeyLiteral = Literal[' ', '!', '+', '-', '*', '=', 't', '?']
+
+
+def is_flagKeyLiteral(inp: str) -> TypeGuard[flagKeyLiteral]:
+    return inp in [' ', '!', '+', '-', '=', '*', 't', '?']
+
 
-flagKeyLiteral = Literal[' ', '!', '+', '-', '*', '=', 't']
 # -------------------------------------------------------------
 
+
 log = logging.getLogger('git.remote')
 log.addHandler(logging.NullHandler())
 
@@ -59,7 +66,9 @@ __all__ = ('RemoteProgress', 'PushInfo', 'FetchInfo', 'Remote')
 #{ Utilities
 
 
-def add_progress(kwargs: Any, git: Git, progress: Union[Callable[..., Any], None]) -> Any:
+def add_progress(kwargs: Any, git: Git,
+                 progress: Union[RemoteProgress, 'UpdateProgress', Callable[..., RemoteProgress], None]
+                 ) -> Any:
     """Add the --progress flag to the given kwargs dict if supported by the
     git command. If the actual progress in the given progress instance is not
     given, we do not request any progress
@@ -75,17 +84,17 @@ def add_progress(kwargs: Any, git: Git, progress: Union[Callable[..., Any], None
 #} END utilities
 
 
-@overload
+@ overload
 def to_progress_instance(progress: None) -> RemoteProgress:
     ...
 
 
-@overload
+@ overload
 def to_progress_instance(progress: Callable[..., Any]) -> CallableRemoteProgress:
     ...
 
 
-@overload
+@ overload
 def to_progress_instance(progress: RemoteProgress) -> RemoteProgress:
     ...
 
@@ -107,7 +116,7 @@ def to_progress_instance(progress: Union[Callable[..., Any], RemoteProgress, Non
     return progress
 
 
-class PushInfo(object):
+class PushInfo(IterableObj, object):
     """
     Carries information about the result of a push operation of a single head::
 
@@ -123,6 +132,7 @@ class PushInfo(object):
         info.summary    # summary line providing human readable english text about the push
         """
     __slots__ = ('local_ref', 'remote_ref_string', 'flags', '_old_commit_sha', '_remote', 'summary')
+    _id_attribute_ = 'pushinfo'
 
     NEW_TAG, NEW_HEAD, NO_MATCH, REJECTED, REMOTE_REJECTED, REMOTE_FAILURE, DELETED, \
         FORCED_UPDATE, FAST_FORWARD, UP_TO_DATE, ERROR = [1 << x for x in range(11)]
@@ -146,11 +156,11 @@ class PushInfo(object):
         self._old_commit_sha = old_commit
         self.summary = summary
 
-    @property
-    def old_commit(self) -> Union[str, SymbolicReference, 'Commit', 'TagObject', 'Blob', 'Tree', None]:
+    @ property
+    def old_commit(self) -> Union[str, SymbolicReference, Commit_ish, None]:
         return self._old_commit_sha and self._remote.repo.commit(self._old_commit_sha) or None
 
-    @property
+    @ property
     def remote_ref(self) -> Union[RemoteReference, TagReference]:
         """
         :return:
@@ -166,7 +176,7 @@ class PushInfo(object):
             raise ValueError("Could not handle remote ref: %r" % self.remote_ref_string)
         # END
 
-    @classmethod
+    @ classmethod
     def _from_line(cls, remote: 'Remote', line: str) -> 'PushInfo':
         """Create a new PushInfo instance as parsed from line which is expected to be like
             refs/heads/master:refs/heads/master 05d2687..1d0568e as bytes"""
@@ -183,7 +193,7 @@ class PushInfo(object):
         # from_to handling
         from_ref_string, to_ref_string = from_to.split(':')
         if flags & cls.DELETED:
-            from_ref = None  # type: Union[SymbolicReference, None]
+            from_ref: Union[SymbolicReference, None] = None
         else:
             if from_ref_string == "(delete)":
                 from_ref = None
@@ -191,7 +201,7 @@ class PushInfo(object):
                 from_ref = Reference.from_path(remote.repo, from_ref_string)
 
         # commit handling, could be message or commit info
-        old_commit = None    # type: Optional[str]
+        old_commit: Optional[str] = None
         if summary.startswith('['):
             if "[rejected]" in summary:
                 flags |= cls.REJECTED
@@ -219,8 +229,13 @@ class PushInfo(object):
 
         return PushInfo(flags, from_ref, to_ref_string, remote, old_commit, summary)
 
+    @ classmethod
+    def iter_items(cls, repo: 'Repo', *args: Any, **kwargs: Any
+                   ) -> NoReturn:  # -> Iterator['PushInfo']:
+        raise NotImplementedError
 
-class FetchInfo(object):
+
+class FetchInfo(IterableObj, object):
 
     """
     Carries information about the results of a fetch operation of a single head::
@@ -237,22 +252,23 @@ class FetchInfo(object):
      info.remote_ref_path # The path from which we fetched on the remote. It's the remote's version of our info.ref
     """
     __slots__ = ('ref', 'old_commit', 'flags', 'note', 'remote_ref_path')
+    _id_attribute_ = 'fetchinfo'
 
     NEW_TAG, NEW_HEAD, HEAD_UPTODATE, TAG_UPDATE, REJECTED, FORCED_UPDATE, \
         FAST_FORWARD, ERROR = [1 << x for x in range(8)]
 
     _re_fetch_result = re.compile(r'^\s*(.) (\[?[\w\s\.$@]+\]?)\s+(.+) -> ([^\s]+)(    \(.*\)?$)?')
 
-    _flag_map = {
+    _flag_map: Dict[flagKeyLiteral, int] = {
         '!': ERROR,
         '+': FORCED_UPDATE,
         '*': 0,
         '=': HEAD_UPTODATE,
         ' ': FAST_FORWARD,
         '-': TAG_UPDATE,
-    }  # type: Dict[flagKeyLiteral, int]
+    }
 
-    @classmethod
+    @ classmethod
     def refresh(cls) -> Literal[True]:
         """This gets called by the refresh function (see the top level
         __init__).
@@ -277,7 +293,7 @@ class FetchInfo(object):
         return True
 
     def __init__(self, ref: SymbolicReference, flags: int, note: str = '',
-                 old_commit: Union['Commit', TagReference, 'Tree', 'Blob', None] = None,
+                 old_commit: Union[Commit_ish, None] = None,
                  remote_ref_path: Optional[PathLike] = None) -> None:
         """
         Initialize a new instance
@@ -291,25 +307,25 @@ class FetchInfo(object):
     def __str__(self) -> str:
         return self.name
 
-    @property
+    @ property
     def name(self) -> str:
         """:return: Name of our remote ref"""
         return self.ref.name
 
-    @property
-    def commit(self) -> 'Commit':
+    @ property
+    def commit(self) -> Commit_ish:
         """:return: Commit of our remote ref"""
         return self.ref.commit
 
-    @classmethod
+    @ classmethod
     def _from_line(cls, repo: 'Repo', line: str, fetch_line: str) -> 'FetchInfo':
         """Parse information from the given line as returned by git-fetch -v
         and return a new FetchInfo object representing this information.
 
-        We can handle a line as follows
-        "%c %-*s %-*s -> %s%s"
+        We can handle a line as follows:
+        "%c %-\\*s %-\\*s -> %s%s"
 
-        Where c is either ' ', !, +, -, *, or =
+        Where c is either ' ', !, +, -, \\*, or =
         ! means error
         + means success forcing update
         - means a tag was updated
@@ -324,8 +340,9 @@ class FetchInfo(object):
             raise ValueError("Failed to parse line: %r" % line)
 
         # parse lines
+        remote_local_ref_str: str
         control_character, operation, local_remote_ref, remote_local_ref_str, note = match.groups()
-        control_character = cast(flagKeyLiteral, control_character)  # can do this neater once 3.5 dropped
+        assert is_flagKeyLiteral(control_character), f"{control_character}"
 
         try:
             _new_hex_sha, _fetch_operation, fetch_note = fetch_line.split("\t")
@@ -342,7 +359,7 @@ class FetchInfo(object):
         # END control char exception handling
 
         # parse operation string for more info - makes no sense for symbolic refs, but we parse it anyway
-        old_commit = None  # type: Union[Commit, TagReference, Tree, Blob, None]
+        old_commit: Union[Commit_ish, None] = None
         is_tag_operation = False
         if 'rejected' in operation:
             flags |= cls.REJECTED
@@ -365,7 +382,7 @@ class FetchInfo(object):
         # If we do not specify a target branch like master:refs/remotes/origin/master,
         # the fetch result is stored in FETCH_HEAD which destroys the rule we usually
         # have. In that case we use a symbolic reference which is detached
-        ref_type = None
+        ref_type: Optional[Type[SymbolicReference]] = None
         if remote_local_ref_str == "FETCH_HEAD":
             ref_type = SymbolicReference
         elif ref_type_name == "tag" or is_tag_operation:
@@ -394,14 +411,15 @@ class FetchInfo(object):
             # by the 'ref/' prefix. Otherwise even a tag could be in refs/remotes, which is when it will have the
             # 'tags/' subdirectory in its path.
             # We don't want to test for actual existence, but try to figure everything out analytically.
-            ref_path = None  # type: Optional[PathLike]
+            ref_path: Optional[PathLike] = None
             remote_local_ref_str = remote_local_ref_str.strip()
+
             if remote_local_ref_str.startswith(Reference._common_path_default + "/"):
                 # always use actual type if we get absolute paths
                 # Will always be the case if something is fetched outside of refs/remotes (if its not a tag)
                 ref_path = remote_local_ref_str
                 if ref_type is not TagReference and not \
-                   remote_local_ref_str.startswith(RemoteReference._common_path_default + "/"):
+                        remote_local_ref_str.startswith(RemoteReference._common_path_default + "/"):
                     ref_type = Reference
                 # END downgrade remote reference
             elif ref_type is TagReference and 'tags/' in remote_local_ref_str:
@@ -420,8 +438,13 @@ class FetchInfo(object):
 
         return cls(remote_local_ref, flags, note, old_commit, local_remote_ref)
 
+    @ classmethod
+    def iter_items(cls, repo: 'Repo', *args: Any, **kwargs: Any
+                   ) -> NoReturn:  # -> Iterator['FetchInfo']:
+        raise NotImplementedError
 
-class Remote(LazyMixin, Iterable):
+
+class Remote(LazyMixin, IterableObj):
 
     """Provides easy read and write access to a git remote.
 
@@ -439,8 +462,9 @@ class Remote(LazyMixin, Iterable):
 
         :param repo: The repository we are a remote of
         :param name: the name of the remote, i.e. 'origin'"""
-        self.repo = repo  # type: 'Repo'
+        self.repo = repo
         self.name = name
+        self.url: str
 
     def __getattr__(self, attr: str) -> Any:
         """Allows to call this instance like
@@ -459,7 +483,7 @@ class Remote(LazyMixin, Iterable):
     def _config_section_name(self) -> str:
         return 'remote "%s"' % self.name
 
-    def _set_cache_(self, attr: str) -> Any:
+    def _set_cache_(self, attr: str) -> None:
         if attr == "_config_reader":
             # NOTE: This is cached as __getattr__ is overridden to return remote config values implicitly, such as
             # in print(r.pushurl)
@@ -496,7 +520,7 @@ class Remote(LazyMixin, Iterable):
             return False
         # end
 
-    @classmethod
+    @ classmethod
     def iter_items(cls, repo: 'Repo', *args: Any, **kwargs: Any) -> Iterator['Remote']:
         """:return: Iterator yielding Remote objects of the given repository"""
         for section in repo.config_reader("repository").sections():
@@ -548,12 +572,12 @@ class Remote(LazyMixin, Iterable):
         """
         return self.set_url(url, delete=True)
 
-    @property
+    @ property
     def urls(self) -> Iterator[str]:
         """:return: Iterator yielding all configured URL targets on a remote as strings"""
         try:
-            # can replace cast with type assert?
-            remote_details = cast(str, self.repo.git.remote("get-url", "--all", self.name))
+            remote_details = self.repo.git.remote("get-url", "--all", self.name)
+            assert isinstance(remote_details, str)
             for line in remote_details.split('\n'):
                 yield line
         except GitCommandError as ex:
@@ -564,14 +588,16 @@ class Remote(LazyMixin, Iterable):
             #
             if 'Unknown subcommand: get-url' in str(ex):
                 try:
-                    remote_details = cast(str, self.repo.git.remote("show", self.name))
+                    remote_details = self.repo.git.remote("show", self.name)
+                    assert isinstance(remote_details, str)
                     for line in remote_details.split('\n'):
                         if '  Push  URL:' in line:
                             yield line.split(': ')[-1]
                 except GitCommandError as _ex:
                     if any(msg in str(_ex) for msg in ['correct access rights', 'cannot run ssh']):
                         # If ssh is not setup to access this repository, see issue 694
-                        remote_details = cast(str, self.repo.git.config('--get-all', 'remote.%s.url' % self.name))
+                        remote_details = self.repo.git.config('--get-all', 'remote.%s.url' % self.name)
+                        assert isinstance(remote_details, str)
                         for line in remote_details.split('\n'):
                             yield line
                     else:
@@ -579,19 +605,19 @@ class Remote(LazyMixin, Iterable):
             else:
                 raise ex
 
-    @property
-    def refs(self) -> IterableList:
+    @ property
+    def refs(self) -> IterableList[RemoteReference]:
         """
         :return:
             IterableList of RemoteReference objects. It is prefixed, allowing
             you to omit the remote path portion, i.e.::
             remote.refs.master # yields RemoteReference('/refs/remotes/origin/master')"""
-        out_refs = IterableList(RemoteReference._id_attribute_, "%s/" % self.name)
+        out_refs: IterableList[RemoteReference] = IterableList(RemoteReference._id_attribute_, "%s/" % self.name)
         out_refs.extend(RemoteReference.list_items(self.repo, remote=self.name))
         return out_refs
 
-    @property
-    def stale_refs(self) -> IterableList:
+    @ property
+    def stale_refs(self) -> IterableList[Reference]:
         """
         :return:
             IterableList RemoteReference objects that do not have a corresponding
@@ -606,13 +632,13 @@ class Remote(LazyMixin, Iterable):
             as well. This is a fix for the issue described here:
             https://github.com/gitpython-developers/GitPython/issues/260
             """
-        out_refs = IterableList(RemoteReference._id_attribute_, "%s/" % self.name)
+        out_refs: IterableList[RemoteReference] = IterableList(RemoteReference._id_attribute_, "%s/" % self.name)
         for line in self.repo.git.remote("prune", "--dry-run", self).splitlines()[2:]:
             # expecting
             # * [would prune] origin/new_branch
             token = " * [would prune] "
             if not line.startswith(token):
-                raise ValueError("Could not parse git-remote prune result: %r" % line)
+                continue
             ref_name = line.replace(token, "")
             # sometimes, paths start with a full ref name, like refs/tags/foo, see #260
             if ref_name.startswith(Reference._common_path_default + '/'):
@@ -624,7 +650,7 @@ class Remote(LazyMixin, Iterable):
         # END for each line
         return out_refs
 
-    @classmethod
+    @ classmethod
     def create(cls, repo: 'Repo', name: str, url: str, **kwargs: Any) -> 'Remote':
         """Create a new remote to the given repository
         :param repo: Repository instance that is to receive the new remote
@@ -641,7 +667,7 @@ class Remote(LazyMixin, Iterable):
     # add is an alias
     add = create
 
-    @classmethod
+    @ classmethod
     def remove(cls, repo: 'Repo', name: str) -> str:
         """Remove the remote with the given name
         :return: the passed remote name to remove
@@ -681,11 +707,12 @@ class Remote(LazyMixin, Iterable):
         return self
 
     def _get_fetch_info_from_stderr(self, proc: TBD,
-                                    progress: Union[Callable[..., Any], RemoteProgress, None]) -> IterableList:
+                                    progress: Union[Callable[..., Any], RemoteProgress, None]
+                                    ) -> IterableList['FetchInfo']:
         progress = to_progress_instance(progress)
 
         # skip first line as it is some remote info we are not interested in
-        output = IterableList('name')
+        output: IterableList['FetchInfo'] = IterableList('name')
 
         # lines which are no progress are fetch info lines
         # this also waits for the command to finish
@@ -741,7 +768,7 @@ class Remote(LazyMixin, Iterable):
         return output
 
     def _get_push_info(self, proc: TBD,
-                       progress: Union[Callable[..., Any], RemoteProgress, None]) -> IterableList:
+                       progress: Union[Callable[..., Any], RemoteProgress, None]) -> IterableList[PushInfo]:
         progress = to_progress_instance(progress)
 
         # read progress information from stderr
@@ -749,7 +776,7 @@ class Remote(LazyMixin, Iterable):
         # read the lines manually as it will use carriage returns between the messages
         # to override the previous one. This is why we read the bytes manually
         progress_handler = progress.new_message_handler()
-        output = IterableList('push_infos')
+        output: IterableList[PushInfo] = IterableList('push_infos')
 
         def stdout_handler(line: str) -> None:
             try:
@@ -784,8 +811,8 @@ class Remote(LazyMixin, Iterable):
             config.release()
 
     def fetch(self, refspec: Union[str, List[str], None] = None,
-              progress: Union[Callable[..., Any], None] = None,
-              verbose: bool = True, **kwargs: Any) -> IterableList:
+              progress: Union[RemoteProgress, None, 'UpdateProgress'] = None,
+              verbose: bool = True, **kwargs: Any) -> IterableList[FetchInfo]:
         """Fetch the latest changes for this remote
 
         :param refspec:
@@ -819,7 +846,7 @@ class Remote(LazyMixin, Iterable):
 
         kwargs = add_progress(kwargs, self.repo.git, progress)
         if isinstance(refspec, list):
-            args = refspec  # type: Sequence[Optional[str]]  # should need this - check logic for passing None through
+            args: Sequence[Optional[str]] = refspec
         else:
             args = [refspec]
 
@@ -831,8 +858,8 @@ class Remote(LazyMixin, Iterable):
         return res
 
     def pull(self, refspec: Union[str, List[str], None] = None,
-             progress: Union[Callable[..., Any], None] = None,
-             **kwargs: Any) -> IterableList:
+             progress: Union[RemoteProgress, 'UpdateProgress', None] = None,
+             **kwargs: Any) -> IterableList[FetchInfo]:
         """Pull changes from the given branch, being the same as a fetch followed
         by a merge of branch with your local branch.
 
@@ -852,8 +879,8 @@ class Remote(LazyMixin, Iterable):
         return res
 
     def push(self, refspec: Union[str, List[str], None] = None,
-             progress: Union[Callable[..., Any], None] = None,
-             **kwargs: Any) -> IterableList:
+             progress: Union[RemoteProgress, 'UpdateProgress', Callable[..., RemoteProgress], None] = None,
+             **kwargs: Any) -> IterableList[PushInfo]:
         """Push changes from source branch in refspec to target branch in refspec.
 
         :param refspec: see 'fetch' method
@@ -883,7 +910,7 @@ class Remote(LazyMixin, Iterable):
                                   universal_newlines=True, **kwargs)
         return self._get_push_info(proc, progress)
 
-    @property
+    @ property
     def config_reader(self) -> SectionConstraint:
         """
         :return:
@@ -898,7 +925,7 @@ class Remote(LazyMixin, Iterable):
             pass
         # END handle exception
 
-    @property
+    @ property
     def config_writer(self) -> SectionConstraint:
         """
         :return: GitConfigParser compatible object able to write options for this remote.
